@@ -1,8 +1,5 @@
 package Test::CGI::External;
 use 5.006;
-require Exporter;
-@ISA = qw(Exporter);
-@EXPORT_OK = qw//;
 use warnings;
 use strict;
 use utf8;
@@ -13,7 +10,7 @@ use File::Temp 'tempfile';
 use FindBin '$Bin';
 use Test::Builder;
 
-our $VERSION = '0.10_03';
+our $VERSION = '0.19_01';
 
 sub new
 {
@@ -21,7 +18,7 @@ sub new
 
     my $tb = Test::Builder->new ();
     $tester{tb} = $tb;
-    $tester{html_validator} = '/home/ben/bin/validate';
+#    $tester{html_validator} = '/home/ben/bin/validate';
 
     return bless \%tester;
 }
@@ -168,10 +165,10 @@ sub test_if_modified_since
     run_private ($self);
     $self->check_headers_private ($self);
     my $headers = $run_options{headers};
+    $self->do_test ($headers->{status}, "Got a status header");
     $self->do_test ($headers->{status} =~ /304/, "Got 304 status response");
     my $body = $run_options{body};
-    $self->do_test (length ($body) == 0, "No body returned with 304 response");
-    # Restore this to whatever it was, probably "undef".
+    $self->do_test (! defined ($body) || length ($body) == 0, "No body returned with 304 response");
     $ENV{HTTP_IF_MODIFIED_SINCE} = $saved;
     # Restore our precious stuff.
     $self->{run_options} = $saved_run_options;
@@ -329,7 +326,14 @@ sub run_private
     else {
 	$self->note ("There is no query string.");
     }
-    my $request_method = $self->check_request_method ($options->{REQUEST_METHOD});
+
+    my $request_method;
+    if ($options->{no_check_request_method}) {
+	$request_method = $options->{REQUEST_METHOD};
+    }
+    else {
+	$request_method = $self->check_request_method ($options->{REQUEST_METHOD});
+    }
     $self->note ("The request method is '$request_method'.");
     setenv_private ($self, 'REQUEST_METHOD', $request_method);
     my $content_type = $options->{CONTENT_TYPE};
@@ -527,6 +531,15 @@ sub check_http_header_syntax_private
     $self->{run_options}{headers} = \%headers;
 }
 
+# The output is required to have a blank line even if it has no body.
+
+sub check_blank_line
+{
+    my ($self, $output) = @_;
+    my $blank = ($output =~ /\r?\n\r?\n/);
+    $self->{tb}->ok ($blank, "Output contains a blank line");
+}
+
 # Check whether the headers of the CGI output are well-formed.
 
 sub check_headers_private
@@ -541,6 +554,7 @@ sub check_headers_private
 	$self->note ("No output, skipping header tests");
         return;
     }
+    check_blank_line ($self, $output);
     my ($header, $body) = split /\r?\n\r?\n/, $output, 2;
     check_http_header_syntax_private ($self, $header, $verbose);
     if (! $self->{no_check_content}) {
@@ -627,9 +641,84 @@ sub set_no_check_content
     $self->{no_check_content} = $value;
 }
 
+sub test_not_implemented
+{
+    my ($self, $method) = @_;
+    my %options;
+    if ($method) {
+	$options{REQUEST_METHOD} = $method;
+    }
+    else {
+	$options{REQUEST_METHOD} = 'GOBBLEDIGOOK';
+    }
+    $options{no_check_request_method} = 1;
+    my $saved_no_check_content = $self->{no_check_content};
+    $self->{no_check_content} = 1;
+    $self->{run_options} = \%options;
+    run_private ($self);
+    #print $options{output}, "\n";
+    $self->check_headers_private ();
+    my $headers = $options{headers};
+    $self->{tb}->ok ($headers->{status}, "Got status header");
+    $self->{tb}->like ($headers->{status}, qr/501/, "Got not implemented status");
+    $self->{no_check_content} = $saved_no_check_content;
+    $self->clear_env ();
+}
+
+sub test_method_not_allowed
+{
+    my ($self, $bad_method) = @_;
+    my $tb = $self->{tb};
+    my %options;
+    $options{REQUEST_METHOD} = $bad_method;
+    $options{no_check_request_method} = 1;
+    my $saved_no_check_content = $self->{no_check_content};
+    $self->{no_check_content} = 1;
+    $self->{run_options} = \%options;
+    run_private ($self);
+    $self->check_headers_private ();
+    my $headers = $options{headers};
+    $tb->ok ($headers->{allow}, "Got Allow header");
+    $tb->like ($headers->{status}, qr/405/, "Got method not allowed status");
+    $self->clear_env ();
+    my @allow = split /,\s*/, $headers->{allow};
+    for my $ok_method (@allow) {
+	# Run the program with each of the headers we were told were
+	# allowed, and see whether the program executes correctly.
+	my %op2;
+	$op2{REQUEST_METHOD} = $ok_method;
+	$self->{run_options} = \%op2;
+	run_private ($self);
+	$self->check_headers_private ();
+	my $headers2 = $op2{headers};
+	# Check that either there is no status line (defaults to 200),
+	# or that there is a status line, and it has status 200.
+	$tb->ok (! $headers2->{status} || $headers2->{status} =~ /200/,
+		 "Method $ok_method specified by Allow: header was allowed");
+	$self->clear_env ();
+    }
+    $self->{no_check_content} = $saved_no_check_content;
+}
+
+# Clear all the environment variables we have set ourselves.
+
+sub clear_env
+{
+my ($self) = @_;
+    for my $e (@{$self->{set_env}}) {
+#        print "Deleting environment variable $e\n";
+        $ENV{$e} = undef;
+    }
+    $self->{set_env} = undef;
+}
+
 sub run
 {
     my ($self, $options) = @_;
+    if (ref $options ne 'HASH') {
+	carp "Use a hash reference as argument, \$tester->run (\\\%options);";
+	return;
+    }
     my $verbose = $self->{verbose};
     local $Test::Builder::Level = $Test::Builder::Level + 1;
     if (! $self->{cgi_executable}) {
@@ -668,6 +757,28 @@ sub run
 	    carp "There is no expected mime type, I suggest text/plain or application/json for JSON output";
 	}
     }
+    elsif ($options->{png} && ! $self->{no_warn}) {
+	my $mime_type = $self->{mime_type};
+	if ($mime_type) {
+	    if ($mime_type ne 'image/png') {
+		carp "Your expected mime type of $mime_type is not valid for PNG";
+	    }
+	}
+	else {
+	    carp "There is no expected mime type, use image/png for PNG output";
+	}
+    }
+
+    if ($options->{png}) {
+	if ($options->{html} || $options->{json}) {
+	    carp "Contradictory options png and json/html";
+	}
+    }
+    elsif ($options->{html}) {
+	if ($options->{json}) {
+	    carp "Contradictory options json and html";
+	}
+    }
 
 #    eval {
     run_private ($self);
@@ -715,11 +826,10 @@ sub run
     if ($options->{json}) {
 	validate_json ($self);
     }
-    for my $e (@{$self->{set_env}}) {
-#        print "Deleting environment variable $e\n";
-        $ENV{$e} = undef;
+    if ($options->{png}) {
+	validate_png ($self);
     }
-    $self->{set_env} = undef;
+    $self->clear_env ();
 }
 
 sub tidy_files
@@ -836,6 +946,10 @@ sub validate_json
 {
     my ($self) = @_;
     my $json = $self->{run_options}->{body};
+    eval "use JSON::Parse 'valid_json';";
+    if ($@) {
+	croak "JSON::Parse is not installed, cannot validate JSON";
+    }
     my $valid = valid_json ($json);
     if ($valid) {
 	$self->pass_test ("Valid JSON");
@@ -845,5 +959,22 @@ sub validate_json
     }
 }
 
+sub validate_png
+{
+    my ($self) = @_;
+    eval "use Image::PNG::Libpng 'read_from_scalar';";
+    if ($@) {
+	croak "Image::PNG::Libpng is not installed, cannot validate PNG";
+    }
+    my $body = $self->{run_options}->{body};
+    my $png;
+    eval {
+	$png = read_from_scalar ($body);
+    };
+    $self->{tb}->ok (!$@, "Could read PNG from body");
+    $self->{tb}->ok ($png, "Got a valid value for PNG");
+    $self->{run_options}{pngdata} = $png;
+}
+
 1;
 
